##############################################################################
#
# DavPack -- A set of patches for all things related to WebDAV
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Zope Corporatioin and
# Contributors.
# The original copyright statement is reproduced below.
#
##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: gzip_pipe_wrapper.py 1114 2006-06-21 15:00:48Z sidnei $
"""

import zlib
import struct
from ZPublisher.HTTPResponse import _gzip_header
from ZServer.Producers import file_part_producer, iterator_producer

COMPRESS_LEVEL = 6
FLUSH_SIZE = 1024 * 4 # 4Kb
CHUNK_TRAILER = '0\r\n\r\n'

class GZipPipeWrapper:

    def __init__(self, pipe, chunked):
        self.pipe = pipe
        self._channel = pipe._channel
        self.chunked = chunked
        self.crc = zlib.crc32("")
        self.zobj = zlib.compressobj(COMPRESS_LEVEL,
                                     zlib.DEFLATED, -zlib.MAX_WBITS,
                                     zlib.DEF_MEM_LEVEL, 0)
        self.write_gzip_header()
        self.size = 0
        self.flush_size = 0

    def write(self, text, l=None):
        if self.chunked and text in (CHUNK_TRAILER,):
            # Ignore premature trailer.
            return
        if not hasattr(text, 'more'):
            # A normal string
            self.write_chunk(text)
        else:
            # It's a producer, we only understand some of them.
            if not isinstance(text, (file_part_producer,
                                     iterator_producer)):
                self.pipe.write(text, l)
                return
            while True:
                chunk = text.more()
                if chunk == '':
                    break
                self.write_chunk(chunk)

    def write_chunk(self, text):
        l = len(text)
        self.size += l
        self.flush_size += l
        self.crc = zlib.crc32(text, self.crc)
        text = self.zobj.compress(text)
        self._write_chunk(text)
        if self.flush_size > FLUSH_SIZE:
            self.sync()
            self.flush_size = 0

    def _write_chunk(self, text):
        l = len(text)
        if not l:
            return
        if self.chunked:
            self.pipe.write('%x\r\n' % l)
        self.pipe.write(text)
        if self.chunked:
            self.pipe.write('\r\n')

    def sync(self):
        self._write_chunk(self.zobj.flush(zlib.Z_SYNC_FLUSH))

    def flush(self):
        self.pipe.flush()

    def finish(self, response):
        self._write_chunk(self.zobj.flush())
        self.write_gzip_trailer(self.crc, self.size)
        self.pipe.finish(response)

    def close(self):
        self.pipe.close()
        self.pipe = None
        self.crc = None
        self.zobj = None
        self.size = None
        self._channel = None

    def write_gzip_header(self):
        self._write_chunk(_gzip_header)

    def write_gzip_trailer(self, crc, size):
        footer = struct.pack("<ll", crc, size)
        self._write_chunk(footer)
        if self.chunked:
            self.pipe.write(CHUNK_TRAILER)
